package de.firstdata.ipg.api.client;

import java.util.Arrays;
import java.util.List;
import java.util.Locale;

import javax.annotation.Nonnull;

import org.eclipse.jdt.annotation.NonNull;

/** XML formatter */
public final class XMLTransformer {

	/** The null character */
	private static final char NULL = '\0';

	/** The space character */
	private static final char SPACE = ' ';

	/** char represents that an XML node started */
	private static final char OPEN_NODE = '<';

	/** char represents that an XML node is being closed */
	private static final char CLOSING_NODE = '/';

	/** char represents that an XML node is now closed */
	private static final char CLOSE_NODE = '>';

	/** char represents that XML prolog started or completed */
	private static final char PROLOG_MARKER = '?';

	/** char for identifying doctype definition or comments */
	private static final char DOCTYPE_COMMENT_MARKER = '!';

	/** a double quote, escaped */
	private static final char ESCAPED_DOUBLE_QUOTE = '\"';

	/** a single quote escaped */
	private static final char ESCAPED_SINGLE_QUOTE = '\'';

	/** Number of space char to add for default indentation */
	private static final int DEFAULT_INDENT = 4;

	/** the character array which forms the line separator */
	@Nonnull
	private static final char[] LINE_BREAK_CHARS = System.lineSeparator().toCharArray();

	/** The prolog string to be added when not available */
	@Nonnull
	private static final String PROLOG = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>";

	/** Name of the nodes whose value must be masked */
	@Nonnull
	private static final List<@NonNull String> NODE_NAME_TO_MASK = Arrays.asList("password", "certificatePassword");

	/**
	 * Define the constructor private, we want only static methods in this class
	 */
	private XMLTransformer() {
		// Utility class
	}

	/**
	 * This method will try to do a single pass of the input string to format it
	 * in pretty XML. For the issues observed, we are trying to avoid any
	 * expensive operations for trying to format the XMLs just for the sake of
	 * printing it in the logs. Formats the XML with default indent of 4
	 *
	 * @param unformattedXML
	 *            - The raw XML with or without line breaks which needs
	 *            formatting
	 * @return formatted XML for pretty printing
	 */
	public static String transform(@Nonnull final String unformattedXML) {
		return transform(unformattedXML, DEFAULT_INDENT);
	}

	/**
	 * This method will try to do a single pass of the input string to format it
	 * in pretty XML. For the issues observed, we are trying to avoid any
	 * expensive operations for trying to format the XMLs just for the sake of
	 * printing it in the logs
	 *
	 * @param unformattedXML
	 *            - The raw XML with or without line breaks which needs
	 *            formatting
	 * @param numOfSpaceForIndents
	 *            - Number of space to be added for indents for pretty printing
	 *            the XML
	 * @return formatted XML for pretty printing
	 */
	public static String transform(@Nonnull final String unformattedXML, final int numOfSpaceForIndents) {
		if (unformattedXML.isEmpty()) {
			return unformattedXML;
		}
		final int length = unformattedXML.length();
		int i = 0;
		// discard any white chars at the beginning of the string, we are just
		// looking for a starting point which should be a not white char. I
		// mean, even a YAML string should start with a non-white char!
		while (i < length - 1 && unformattedXML.charAt(i) < 33) {
			i++;
		}
		// if the input XML string has a prolog, we collect that. Prolog do not
		// need to be intended
		final StringBuilder formattedXML = new StringBuilder(2 * length);
		boolean prologAvailable = false;
		if (OPEN_NODE == unformattedXML.charAt(i) && PROLOG_MARKER == unformattedXML.charAt(i + 1)) {
			char previousChar = NULL;
			while (i < length) {
				final char currentChar = unformattedXML.charAt(i++);
				formattedXML.append(currentChar);
				if (CLOSE_NODE == currentChar && PROLOG_MARKER == previousChar) {
					break;
				}
				previousChar = currentChar;
			}
			for (final char lineSeparatorChar: LINE_BREAK_CHARS) {
				formattedXML.append(lineSeparatorChar);
			}
			// after the prolog, we discard any white char as well
			while (i < length - 1 && unformattedXML.charAt(i) < 33) {
				i++;
			}
			prologAvailable = true;
		}
		// grab the doctype definition or comments, if any. DTD definition need
		// not be indented and copy just as is
		if (OPEN_NODE == unformattedXML.charAt(i) && DOCTYPE_COMMENT_MARKER == unformattedXML.charAt(i + 1)) {
			int openTagCount = 1;
			char currentChar = unformattedXML.charAt(i++);
			formattedXML.append(currentChar);
			while (i < length && openTagCount != 0) {
				currentChar = unformattedXML.charAt(i++);
				formattedXML.append(currentChar);
				if (ESCAPED_DOUBLE_QUOTE != currentChar && ESCAPED_SINGLE_QUOTE != currentChar) {
					if (OPEN_NODE == currentChar) {
						openTagCount++;
					} else if (CLOSE_NODE == currentChar) {
						openTagCount--;
					}
				}
			}
			for (final char lineSeparatorChar: LINE_BREAK_CHARS) {
				formattedXML.append(lineSeparatorChar);
			}
			// after the doctype definition, we discard any white chars
			while (i < length - 1 && unformattedXML.charAt(i) < 33) {
				i++;
			}
		}
		char previousPreviousChar = NULL;
		char currentChar = unformattedXML.charAt(i++);
		char previousChar = currentChar;
		boolean newNode = OPEN_NODE == currentChar;
		// at this point we hope the actual XML body starts, so append if a new
		// node started
		if (newNode) {
			formattedXML.append(currentChar);
		}
		int indentCount = 0;
		int nodeNameStarts = 0;
		boolean mask = false;
		final Locale locale = Locale.getDefault();
		while (i < length - 1) {
			currentChar = unformattedXML.charAt(i++);
			// discard all white chars when a node is not already started. if a
			// node is started we read all chars in it - CDATA, PCDATA, etc.
			if (currentChar < 33 && !newNode) {
				continue;
			}
			final char nextChar = unformattedXML.charAt(i);
			switch (currentChar) {
			case OPEN_NODE:
				mask = false;
				// a new node started
				if (CLOSE_NODE == previousChar && CLOSING_NODE != nextChar && DOCTYPE_COMMENT_MARKER != nextChar) {
					indentCount++;
				}
				if (CLOSING_NODE != nextChar && CLOSE_NODE == previousChar) {
					for (final char lineSeparatorChar: LINE_BREAK_CHARS) {
						formattedXML.append(lineSeparatorChar);
					}
					for (int j = indentCount * numOfSpaceForIndents; j > 0; j--) {
						formattedXML.append(SPACE);
					}
				}
				formattedXML.append(currentChar);
				newNode = CLOSE_NODE == previousChar || DOCTYPE_COMMENT_MARKER != nextChar;
				if (newNode) {
					nodeNameStarts = formattedXML.length();
				}
				break;
			case CLOSE_NODE:
				// a node closes
				formattedXML.append(currentChar);
				char nextNonWhiteChar = unformattedXML.charAt(i++);
				while (nextNonWhiteChar < 33 && i < length - 1) {
					nextNonWhiteChar = unformattedXML.charAt(i++);
				}
				i--;
				newNode = OPEN_NODE != nextNonWhiteChar;
				if (newNode) {
					mask = shouldNodeValueBeMasked(formattedXML.substring(nodeNameStarts, formattedXML.length() - 1).toLowerCase(locale));
					nodeNameStarts = 0;
				}
				break;
			case CLOSING_NODE:
				// anode about to be closed
				if (OPEN_NODE == previousChar || CLOSE_NODE == nextChar) {
					indentCount--;
				}
				if (OPEN_NODE == previousChar && CLOSE_NODE == previousPreviousChar) {
					final int currentLength = formattedXML.length() - 1;
					final char lastChar = formattedXML.charAt(currentLength);
					formattedXML.setCharAt(currentLength, LINE_BREAK_CHARS[0]);
					for (int j = 1; j < LINE_BREAK_CHARS.length; j++) {
						formattedXML.append(LINE_BREAK_CHARS[j]);
					}
					for (int j = (indentCount + 1) * numOfSpaceForIndents; j > 0; j--) {
						formattedXML.append(SPACE);
					}
					formattedXML.append(lastChar);
				}
				formattedXML.append(currentChar);
				break;
			default:
				// grab everything else
				if (mask) {
					formattedXML.append('X');
				} else {
					formattedXML.append(currentChar);
				}
			}
			previousPreviousChar = previousChar;
			previousChar = currentChar;
		}
		formattedXML.append(unformattedXML.charAt(length - 1));
		// it would appear there are some JUnits which are explicitly trying to
		// validate the addition of prolog in the XML object after
		// transformation and hence the immediately below conditional check. To
		// keep things simple, we add prolog for SOAP message as it would still
		// be a valid SOAP message
		if (!prologAvailable) {
			return PROLOG + System.lineSeparator() + formattedXML.toString();
		}
		return formattedXML.toString();
	}

	/**
	 * Check if node's value should be masked.
	 *
	 * @param nodeName
	 *            Name of the node
	 * @return {@code true} if should be masked, {@code false} otherwise
	 */
	private static boolean shouldNodeValueBeMasked(@Nonnull final String nodeName) {
		for (final String knownNode: NODE_NAME_TO_MASK) {
			if (nodeName.contains(knownNode)) {
				return true;
			}
		}
		return false;
	}
}
