package de.firstdata.ipg.api.client.exception;

import de.firstdata.ipg.api.client.IPGApiClient;
import de.firstdata.ipg.api.schema.IPGApiOrderResponse;

/**
 * An instance of this class is thrown in case the credit card processor has
 * failed to process your transaction due to invalid transaction data. It
 * contains attributes describing the error cause in detail. You can access
 * these attributes by calling the appropriate <code>getter</code> methods.
 *
 * @author Andreas Schmid
 *
 */
public class ProcessingException extends ClientException {

	private static final long serialVersionUID = 1L;

	private final IPGApiOrderResponse _response;

	/**
	 * creates an instance of this class taking an error message and an
	 * unmarhalled response object as arguments - note that you should never
	 * call this constructor, since it is internally invoked by
	 * {@link IPGApiClient#commitTransaction(de.firstdata.ipgapi.client.IPGApiTransaction)}
	 * when receiving a processing error from the First Data API Web Service
	 *
	 * @param message
	 *            a string describing the error
	 * @param response
	 *            an unmarshalled response object
	 */
	public ProcessingException(final String message, final IPGApiOrderResponse response) {
		super(message);
		this._response = response;
	}

	/**
	 * returns the commetcial service provider
	 *
	 * @return the commercial service provider - currently, <i>BNLP</i> will
	 *         always be returned
	 */
	public String getCommercialServiceProvider() {
		return this._response.getCommercialServiceProvider();
	}

	/**
	 * returns the time stamp set by the First Data API Web Service before
	 * returning the transaction result
	 *
	 * @return the time stamp in millisecs representing the difference between
	 *         the current time and midnight, January 1, 1970 UTC
	 */
	public String getTransactionTime() {
		return this._response.getTransactionTime();
	}

	/**
	 * returns the reference number the credit card processor uses for referring
	 * to this transaction - note that this element might be empty in case the
	 * transaction fas failed before reaching the credit card processor
	 *
	 * @return the reference number as returned by the credit card processor
	 */
	public String getProcessorReferenceNumber() {
		return this._response.getProcessorReferenceNumber();
	}

	/**
	 * returns the processing error message returned by the credit card
	 * processor - note that this element might be empty in case the transaction
	 * fas failed before reaching the credit card processor
	 *
	 * @return the error message returned as by the credit card processor
	 */
	public String getProcessorResponseMessage() {
		return this._response.getProcessorResponseMessage();
	}

	/**
	 * returns the processing error message starting with the error code
	 * followed by an error description - in case the error has been found by
	 * the credit card processor, this description is the same as the one
	 * returned by {@link ProcessingException#getProcessorResponseMessage()}
	 *
	 * @return the processing error message consisting of an error code and an
	 *         error description
	 */
	public String getErrorMessage() {
		return this._response.getErrorMessage();
	}

	/**
	 * returns the order ID with which this transaction can be referred to -
	 * note that this ID cannot be provided for any follow-up transactions since
	 * the transaction has failed
	 *
	 * @return the transaction's order ID as returned by the First Data API Web
	 *         Service
	 */
	public String getOrderId() {
		return this._response.getOrderId();
	}

	/**
	 * returns the transaction ID with which this transaction can be referred to
	 * - note that this ID cannot be provided for any follow-up transactions
	 * since the transaction has failed
	 *
	 * @return the transaction ID as returned by the First Data API Web Service
	 */
	public Long getIpgTransactionId() {
		return this._response.getIpgTransactionId();
	}

	/**
	 * returns the code which is created by the credit card processor when
	 * having processing this transaction - note that this element might be
	 * empty in case the transaction fas failed before reaching the credit card
	 * processor
	 *
	 * @return the approval code as returned by the credit card processor
	 */
	public String getApprovalCode() {
		return this._response.getApprovalCode();
	}

	/**
	 * returns the address verification response (AVS) - currently, AVS is not
	 * supported by the First Data API, i.e. an empty string is always returned
	 *
	 * @return an empty string
	 */
	public String getAVSResponse() {
		return this._response.getAVSResponse();
	}

	/**
	 * returns the tDate of this transaction - note that this tDate cannot be
	 * used when voiding the transaction since it has failed
	 *
	 * @return the transaction's tDate as returned by the First Data API Web
	 *         Service
	 */
	public String getTDate() {
		return this._response.getTDate();
	}

	/**
	 * returns the transaction result which is either <i>DECLINE</i> or
	 * <i>FRAUD</i> depending on the error cause
	 *
	 * @return either the string <i>DECLINE</i> or <i>FRAUD</i>
	 */
	public String getTransactionResult() {
		return this._response.getTransactionResult();
	}

	public String getProcessorResponseCode() {
		return this._response.getProcessorResponseCode();
	}

	public IPGApiOrderResponse getResponse() {
		return this._response;
	}

	public String getExternalMerchantID() {
		return getResponse().getExternalMerchantID();
	}

	public String getBrand() {
		return getResponse().getBrand();
	}
}
